package de.tudarmstadt.informatik.hostage.protocol;

import java.util.ArrayList;
import java.util.List;

import de.tudarmstadt.informatik.hostage.wrapper.ByteArray;

public final class TELNET implements Protocol<ByteArray> {

	private static enum STATE {
		NONE, OPEN, CLOSED, LOGIN, AUTHENTICATE, LOGGED_IN
	};

	private STATE state = STATE.NONE;

	private byte[] usr;
	private byte[] cmd;
	private String machine = "raspberrypi";
	private byte[] session = null;

	@Override
	public int getPort() {
		return 23;
	}

	@Override
	public TALK_FIRST whoTalksFirst() {
		return TALK_FIRST.SERVER;
	}

	@Override
	public List<ByteArray> processMessage(ByteArray message) {
		byte[] request = message.get();
		List<ByteArray> response = new ArrayList<ByteArray>();
		switch (state) {
		case NONE:
			response.add(new ByteArray(cmdRequests));
			state = STATE.OPEN;
			break;
		case OPEN:
			if(request != null) {
				response.add(new ByteArray(getCmdResponses(request)));
				response.add(new ByteArray("Debian GNU/Linux 7.0\r\n"));
				response.add(new ByteArray(machine + "login: "));
				state = STATE.LOGIN;
			}
			break;
		case LOGIN:
			if(request == null) break;
			else if(request[0] == 0x0d) {
				response.add(new ByteArray("\r\n"));
				response.add(new ByteArray("Password: "));
				state = STATE.AUTHENTICATE;
				session = concat(sessionPrefix, "@".getBytes(), machine.getBytes(), sessionMiddle, "@".getBytes(), machine.getBytes(), sessionSuffix);
				break;
			} else if (request[0] == 0x7f) {
				byte[] tmp = new byte[usr.length - 1];
				System.arraycopy(usr, 0, tmp, 0, usr.length - 1);
				usr = tmp;
				response.add(new ByteArray("\b \b"));
			} else if (request[0] != (byte) 0xff){
				if(usr == null)
					usr = request;
				else
					usr = concat(usr, request);
				response.add(message);
			}
			break;
		case AUTHENTICATE:
			if(request == null) break;
			else if(request[0] == 0x0d) {
				response.add(new ByteArray("Last Login: \r\nLinux" + machine + " 3.6.11+\r\n"));
				response.add(new ByteArray(session));
				state = STATE.LOGGED_IN;
			} else if (request[0] == 0x7f) {
				response.add(new ByteArray("\b \b"));
			}			
			break;
		case LOGGED_IN:
			if(request == null) break;
			else if(request[0] == 0x0d) {
				if(cmd == null) {
					response.add(new ByteArray("\r\n"));
					response.add(new ByteArray(session));
				} else if (new String(cmd).contains("exit")) {
					response.add(new ByteArray("\r\nlogout\r\n"));
					state = STATE.CLOSED;
				} else {
					String bash = "\r\n-bash: " + new String(cmd)+ ": command not found";
					response.add(new ByteArray(bash));
					response.add(new ByteArray("\r\n"));
					response.add(new ByteArray(session));
					cmd = null;
				}
			} else if (request[0] == 0x7f) {
				byte[] tmp = new byte[cmd.length - 1];
				System.arraycopy(cmd, 0, tmp, 0, cmd.length - 1);
				cmd = tmp;
				response.add(new ByteArray("\b \b"));
			} else if (request[0] != (byte) 0xff){
				if(cmd == null)
					cmd = request;
				else
					cmd = concat(cmd, request);
				response.add(message);
			}
			break;
		default:
			response.add(new ByteArray("\r\nlogout\r\n"));
			state = STATE.CLOSED;
			break;
		}
		return response;
	}

	@Override
	public boolean isClosed() {
		return (state == STATE.CLOSED);
	}

	@Override
	public boolean isSecure() {
		return false;
	}
	@Override
	public String toString() {
		return "TELNET";
	}
	
	@Override
	public Class<ByteArray> getType() {
		return ByteArray.class;
	}

	private byte[] getCmdResponses(byte[] request) {
		List<byte[]> respList = new ArrayList<byte[]>();
		byte[] cmdResp;
		for(int i = 0; i < request.length - 2; i += 3) {
			if(request[i] == (byte) 0xff && request[i+2] != 0x03 && request[i+2] != 0x01) {
				cmdResp = new byte[3];
				cmdResp[0] = request[i];
				cmdResp[1] = request[i + 1] == (byte) 0xfd ? (byte) 0xfc : (byte) 0xfe; 
				cmdResp[2] = request[i+2];
				respList.add(cmdResp);
			}			
		}
		byte[] response = new byte[0];
		for(byte[] resp : respList) {
			response = concat(response, resp);
		}
		return response;
	}

	private final byte[] cmdRequests = {(byte) 0xff, (byte) 0xfb, 0x03,
										(byte) 0xff, (byte) 0xfb, 0x01};

	private final byte[] sessionPrefix = {0x1b, 0x5d, 0x30, 0x3b};
	private final byte[] sessionMiddle = {0x40, 0x72, 0x61, 0x73, 
			0x70, 0x62, 0x65, 0x72, 0x72, 0x79, 0x70, 0x69, 0x3a, 0x20, 0x7e, 0x07, 0x1b, 0x5b, 0x30, 0x31, 
			0x3b, 0x33, 0x32, 0x6d};	
	private final byte[] sessionSuffix = {0x1b, 0x5b, 0x30, 0x30, 0x6d, 0x20, 0x1b, 0x5b, 0x30, 0x31, 
			0x3b, 0x33, 0x34, 0x6d, 0x7e, 0x20, 0x24, 0x1b, 0x5b, 0x30, 0x30, 0x6d, 0x20};

	private byte[] concat(byte[]... bytes) {
		int newSize = 0;
		for (byte[] b : bytes)
			newSize += b.length;
		byte[] dst = new byte[newSize];

		int currentPos = 0;
		int newPos;
		for (byte[] b : bytes) {
			newPos = b.length;
			System.arraycopy(b, 0, dst, currentPos, newPos);
			currentPos += newPos;
		}
		return dst;
	}

}